package org.xwiki.cache.jbosscache.internal;

import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.jboss.cache.notifications.annotation.NodeModified;
import org.jboss.cache.notifications.event.NodeModifiedEvent;
import org.jboss.cache.Cache;
import org.jboss.cache.CacheFactory;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.Node;
import org.xwiki.cache.DisposableCacheValue;
import org.xwiki.cache.config.CacheConfiguration;
import org.xwiki.cache.jbosscache.internal.event.JBossCacheCacheEntryEvent;
import org.xwiki.cache.util.AbstractCache;
import org.xwiki.container.Container;

public class JBossCacheCache<T> extends AbstractCache<T> implements org.xwiki.cache.Cache<T>
{
    public static final Fqn<String> ROOT_FQN = Fqn.fromString("/xwiki");

    private static final Log LOG = LogFactory.getLog(JBossCacheCache.class);

    private static final String DATA_KEY = "data";

    Container container;

    protected String defaultPropsId = "default";

    private JBossCacheCacheConfiguration configuration;

    private Cache<String, T> cache;

    private Node<String, T> rootNode;

    private Map<String, T> preEventData;

    public JBossCacheCache()
    {

    }

    public void initialize(CacheConfiguration configuration)
    {
        this.configuration = new JBossCacheCacheConfiguration(container, configuration, defaultPropsId);

        CacheFactory<String, T> factory = new DefaultCacheFactory<String, T>();
        this.cache = factory.createCache(this.configuration.getJBossCacheConfiguration());

        this.cache.addCacheListener(ROOT_FQN, this);

        this.cache.create();
        this.cache.start();

        this.rootNode = cache.getNode(ROOT_FQN);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#remove(java.lang.String)
     */
    public void remove(String key)
    {
        this.rootNode.removeChild(key);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#set(java.lang.String, java.lang.Object)
     */
    public void set(String key, T obj)
    {
        this.rootNode.getChild(key).put(DATA_KEY, obj);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#get(java.lang.String)
     */
    public T get(String key)
    {
        return this.rootNode.getChild(key).get(DATA_KEY);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#getSize()
     */
    public int getSize()
    {
        return this.rootNode.getChildren().size();
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#removeAll()
     */
    public void removeAll()
    {
        this.cache.removeNode(ROOT_FQN);
        this.rootNode = cache.getNode(ROOT_FQN);
    }

    public void dispose()
    {
        super.dispose();

        this.cache.stop();
        this.cache.destroy();
    }

    private void tryDisposeEntryValue(T cacheValue)
    {
        if (cacheValue instanceof DisposableCacheValue) {
            try {
                ((DisposableCacheValue) cacheValue).dispose();
            } catch (Throwable e) {
                if (LOG.isWarnEnabled()) {
                    LOG.warn("Error when trying to dispose a cache object of cache ["
                        + configuration.getCacheConfiguration().getConfigurationId() + "]", e);
                }
            }
        }
    }

    // ////////////////////////////////////////////////////////////////
    // Events
    // ////////////////////////////////////////////////////////////////

    @NodeModified
    public void nodeModified(NodeModifiedEvent event)
    {
        Map<String, T> data = event.getData();

        if (event.isPre()) {
            this.preEventData = data;
        } else {
            if (data.containsKey(DATA_KEY)) {
                String key = event.getFqn().getLastElementAsString();
                T value = data.get(DATA_KEY);

                if (event.getModificationType() == NodeModifiedEvent.ModificationType.REMOVE_DATA) {
                    cacheEntryRemoved(key, value);
                } else {
                    cacheEntryAdded(key, value);
                }
            }

            this.preEventData = null;
        }
    }

    private void cacheEntryAdded(String key, T value)
    {
        JBossCacheCacheEntryEvent<T> event =
            new JBossCacheCacheEntryEvent<T>(new JBossCacheCacheEntry<T>(this, key, value));

        for (org.xwiki.cache.event.CacheEntryListener<T> listener : this.cacheEntryListeners
            .getListeners(org.xwiki.cache.event.CacheEntryListener.class)) {

            if (this.preEventData.containsKey(key)) {
                listener.cacheEntryModified(event);
            } else {
                listener.cacheEntryAdded(event);
            }
        }
    }

    private void cacheEntryRemoved(String key, T value)
    {
        JBossCacheCacheEntryEvent<T> event =
            new JBossCacheCacheEntryEvent<T>(new JBossCacheCacheEntry<T>(this, key, value));

        for (org.xwiki.cache.event.CacheEntryListener<T> listener : this.cacheEntryListeners
            .getListeners(org.xwiki.cache.event.CacheEntryListener.class)) {
            listener.cacheEntryRemoved(event);
        }

        tryDisposeEntryValue(value);
    }
}
