package org.xwiki.cache.jbosscache.internal;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.UUID;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.EvictionConfig;
import org.jboss.cache.config.EvictionPolicyConfig;
import org.jboss.cache.config.EvictionRegionConfig;
import org.jboss.cache.eviction.LRUConfiguration;
import org.jboss.cache.eviction.LRUPolicy;
import org.jboss.cache.factories.XmlConfigurationParser;
import org.xwiki.cache.config.CacheConfiguration;
import org.xwiki.cache.eviction.EntryEvictionConfiguration;
import org.xwiki.cache.eviction.LRUEvictionConfiguration;
import org.xwiki.container.Container;

public class JBossCacheCacheConfiguration
{
    private static final Log LOG = LogFactory.getLog(JBossCacheCacheConfiguration.class);

    private static final String PROPS_PATH = "cache/jboss/";

    private static final String PROPS_EXT = ".xml";

    protected String defaultPropsId = "default";

    private CacheConfiguration configuration;

    private Configuration jbossConfiguration;

    private Container container;

    private int timeToLive = -1;

    private String name;

    public JBossCacheCacheConfiguration(Container container, CacheConfiguration configuration, String defaultPropsId)
    {
        this.container = container;
        this.configuration = configuration;
        this.defaultPropsId = defaultPropsId;

        load();
    }

    public CacheConfiguration getCacheConfiguration()
    {
        return this.configuration;
    }

    public Configuration getJBossCacheConfiguration()
    {
        return this.jbossConfiguration;
    }

    public int getTimeToLive()
    {
        return this.timeToLive;
    }

    public String getName()
    {
        return this.name;
    }

    private void load()
    {
        this.jbossConfiguration = loadConfig(configuration.getConfigurationId());

        if (this.jbossConfiguration == null) {
            EntryEvictionConfiguration eec =
                (EntryEvictionConfiguration) this.configuration.get(EntryEvictionConfiguration.CONFIGURATIONID);

            if (eec != null) {
                EvictionConfig ec = this.jbossConfiguration.getEvictionConfig();

                if (eec.getMode() == EntryEvictionConfiguration.Mode.LRU) {
                    int maxEntries = ((Number) eec.get(LRUEvictionConfiguration.MAXENTRIES_ID)).intValue();

                    ec.setDefaultEvictionPolicyClass(LRUPolicy.class.getName());

                    LRUConfiguration lru = new LRUConfiguration();

                    lru.setMaxNodes(maxEntries);

                    if (eec.getTimeToLive() > 0) {
                        lru.setTimeToLiveSeconds(eec.getTimeToLive() / 1000);
                    }

                    setEvictionConfiguration(ec, eec, lru);
                }
            }

        }

        this.jbossConfiguration.setClusterName(this.configuration.getConfigurationId());
    }

    protected Configuration getDefaultConfig()
    {
        return loadConfig(defaultPropsId);
    }

    protected Configuration getConfig(CacheConfiguration config)
    {
        if (LOG.isInfoEnabled()) {
            LOG.info("Loading cache properties: " + config.getConfigurationId());
        }

        Configuration properties = loadConfig(config.getConfigurationId());

        if (config == null) {
            EntryEvictionConfiguration eec =
                (EntryEvictionConfiguration) config.get(EntryEvictionConfiguration.CONFIGURATIONID);

            if (eec != null) {
                EvictionConfig ec = properties.getEvictionConfig();

                if (eec.getMode() == EntryEvictionConfiguration.Mode.LRU) {
                    int maxEntries = ((Number) eec.get(LRUEvictionConfiguration.MAXENTRIES_ID)).intValue();

                    ec.setDefaultEvictionPolicyClass(LRUPolicy.class.getName());

                    LRUConfiguration lru = new LRUConfiguration();

                    lru.setMaxNodes(maxEntries);

                    if (eec.getTimeToLive() > 0) {
                        lru.setTimeToLiveSeconds(eec.getTimeToLive() / 1000);
                    }

                    setEvictionConfiguration(ec, eec, lru);
                }
            }

        }

        String name = this.configuration.getConfigurationId();

        if (name == null) {
            name = UUID.randomUUID().toString();
        }

        properties.setClusterName(name);

        return properties;
    }

    private void setEvictionConfiguration(EvictionConfig ec, EntryEvictionConfiguration eec, EvictionPolicyConfig epc)
    {
        EvictionRegionConfig erc = new EvictionRegionConfig();

        erc.setRegionFqn(JBossCacheCache.ROOT_FQN);

        erc.setEvictionPolicyConfig(epc);

        ec.setEvictionRegionConfigs(Collections.singletonList(erc));
    }

    private Configuration loadConfig(String propertiesId)
    {
        Configuration config = null;

        String propertiesFilename = propertiesId + PROPS_EXT;

        File file = new File(PROPS_PATH + propertiesFilename);

        InputStream is = null;

        try {
            if (file.exists()) {
                is = new FileInputStream(file);
            } else {
                is =
                    this.container.getApplicationContext().getResourceAsStream(
                        "/WEB-INF/" + PROPS_PATH + propertiesFilename);
            }

            XmlConfigurationParser parser = new XmlConfigurationParser();
            config = parser.parseStream(is);

            if (LOG.isInfoEnabled()) {
                LOG.info("Properties loaded: " + propertiesFilename);
            }
        } catch (FileNotFoundException e) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("Failed to load configuration file " + propertiesId, e);
            }
        } finally {
            if (is != null) {
                try {
                    is.close();
                } catch (IOException e) {
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Failed t close properties file", e);
                    }
                }
            }
        }

        return config;
    }
}
