package org.xwiki.cache.oscache.internal;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xwiki.cache.DisposableCacheValue;
import org.xwiki.cache.config.CacheConfiguration;
import org.xwiki.cache.oscache.internal.event.OSCacheCacheEntryEvent;
import org.xwiki.cache.util.AbstractCache;
import org.xwiki.container.Container;

import com.opensymphony.oscache.base.NeedsRefreshException;
import com.opensymphony.oscache.base.events.CacheEntryEvent;
import com.opensymphony.oscache.base.events.CacheEntryEventListener;
import com.opensymphony.oscache.base.events.CacheGroupEvent;
import com.opensymphony.oscache.base.events.CachePatternEvent;
import com.opensymphony.oscache.base.events.CachewideEvent;
import com.opensymphony.oscache.general.GeneralCacheAdministrator;

public class OSCacheCache<T> extends AbstractCache<T> implements org.xwiki.cache.Cache<T>, CacheEntryEventListener
{
    private static final Log LOG = LogFactory.getLog(OSCacheCache.class);

    Container container;

    protected String defaultPropsId = "default";
    
    private OSCacheCacheConfiguration config;

    private GeneralCacheAdministrator cacheAdmin;

    public OSCacheCache()
    {
    }

    public void initialize(CacheConfiguration configuration)
    {
        this.config = new OSCacheCacheConfiguration(this.container, configuration, this.defaultPropsId);

        this.cacheAdmin = new GeneralCacheAdministrator(this.config.getOSCacheProperties());

        this.cacheAdmin.getCache().addCacheEventListener(this);
    }

    public String getName()
    {
        return this.config.getName();
    }

    private String cacheKey(String key)
    {
        return getName() + key;
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#remove(java.lang.String)
     */
    public void remove(String key)
    {
        this.cacheAdmin.flushEntry(cacheKey(key));
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#set(java.lang.String, java.lang.Object)
     */
    public void set(String key, T obj)
    {
        this.cacheAdmin.putInCache(cacheKey(key), obj);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#get(java.lang.String)
     */
    public T get(String key)
    {
        T value = null;

        try {
            value = (T) this.cacheAdmin.getFromCache(cacheKey(key), config.getTimeToLive());
        } catch (NeedsRefreshException e) {
            this.cacheAdmin.cancelUpdate(key);

            if (LOG.isDebugEnabled()) {
                LOG.debug("Failed to get the value from the cache", e);
            }
        }

        return value;
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#removeAll()
     */
    public void removeAll()
    {
        this.cacheAdmin.flushAll();
    }

    /**
     * {@inheritDoc}
     * <p>
     * Method called here is labeled "for test only" in OSCache JavaDoc, so don't call this a lot!
     * </p>
     * 
     * @see org.xwiki.cache.Cache#getSize()
     */
    public int getSize()
    {
        return this.cacheAdmin.getCache().getSize();
    }

    public void dispose()
    {
        super.dispose();

        this.cacheAdmin.destroy();
    }

    private void tryDisposeEntryValue(T cacheValue)
    {
        if (cacheValue instanceof DisposableCacheValue) {
            try {
                ((DisposableCacheValue) cacheValue).dispose();
            } catch (Throwable e) {
                if (LOG.isWarnEnabled()) {
                    LOG.warn("Error when trying to dispose a cache object of cache ["
                        + config.getCacheConfiguration().getConfigurationId() + "]", e);
                }
            }
        }
    }

    // ////////////////////////////////////////////////////////////////
    // Events
    // ////////////////////////////////////////////////////////////////

    /**
     * {@inheritDoc}
     * 
     * @see com.opensymphony.oscache.base.events.CacheEntryEventListener#cacheEntryAdded(com.opensymphony.oscache.base.events.CacheEntryEvent)
     */
    public void cacheEntryAdded(CacheEntryEvent event)
    {
        for (org.xwiki.cache.event.CacheEntryListener<T> listener : cacheEntryListeners
            .getListeners(org.xwiki.cache.event.CacheEntryListener.class)) {
            listener.cacheEntryAdded(new OSCacheCacheEntryEvent<T>(this, event));
        }
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.opensymphony.oscache.base.events.CacheEntryEventListener#cacheEntryFlushed(com.opensymphony.oscache.base.events.CacheEntryEvent)
     */
    public void cacheEntryFlushed(CacheEntryEvent event)
    {
        for (org.xwiki.cache.event.CacheEntryListener<T> listener : cacheEntryListeners
            .getListeners(org.xwiki.cache.event.CacheEntryListener.class)) {
            listener.cacheEntryRemoved(new OSCacheCacheEntryEvent<T>(this, event));
        }

        tryDisposeEntryValue((T) event.getEntry().getContent());
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.opensymphony.oscache.base.events.CacheEntryEventListener#cacheEntryRemoved(com.opensymphony.oscache.base.events.CacheEntryEvent)
     */
    public void cacheEntryRemoved(CacheEntryEvent event)
    {
        for (org.xwiki.cache.event.CacheEntryListener<T> listener : cacheEntryListeners
            .getListeners(org.xwiki.cache.event.CacheEntryListener.class)) {
            listener.cacheEntryRemoved(new OSCacheCacheEntryEvent<T>(this, event));
        }

        tryDisposeEntryValue((T) event.getEntry().getContent());
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.opensymphony.oscache.base.events.CacheEntryEventListener#cacheEntryUpdated(com.opensymphony.oscache.base.events.CacheEntryEvent)
     */
    public void cacheEntryUpdated(CacheEntryEvent event)
    {
        for (org.xwiki.cache.event.CacheEntryListener<T> listener : cacheEntryListeners
            .getListeners(org.xwiki.cache.event.CacheEntryListener.class)) {
            listener.cacheEntryModified(new OSCacheCacheEntryEvent<T>(this, event));
        }
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.opensymphony.oscache.base.events.CacheEntryEventListener#cacheGroupFlushed(com.opensymphony.oscache.base.events.CacheGroupEvent)
     */
    public void cacheGroupFlushed(CacheGroupEvent event)
    {
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.opensymphony.oscache.base.events.CacheEntryEventListener#cachePatternFlushed(com.opensymphony.oscache.base.events.CachePatternEvent)
     */
    public void cachePatternFlushed(CachePatternEvent event)
    {
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.opensymphony.oscache.base.events.CacheEntryEventListener#cacheFlushed(com.opensymphony.oscache.base.events.CachewideEvent)
     */
    public void cacheFlushed(CachewideEvent event)
    {
    }
}
