package org.xwiki.cache.oscache.internal;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;
import java.util.Properties;
import java.util.UUID;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xwiki.cache.config.CacheConfiguration;
import org.xwiki.cache.eviction.EntryEvictionConfiguration;
import org.xwiki.cache.eviction.LRUEvictionConfiguration;
import org.xwiki.container.Container;

import com.opensymphony.oscache.base.algorithm.LRUCache;

public class OSCacheCacheConfiguration
{
    private static final Log LOG = LogFactory.getLog(OSCacheCacheConfiguration.class);

    private static final String PROPS_PATH = "cache/oscache/";

    private static final String PROPS_EXT = ".properties";

    protected String defaultPropsId = "default";

    private CacheConfiguration configuration;

    private Properties oscacheConfiguration;

    private Container container;

    private int timeToLive = -1;

    private String name;

    public OSCacheCacheConfiguration(Container container, CacheConfiguration configuration, String defaultPropsId)
    {
        this.container = container;
        this.configuration = configuration;
        this.defaultPropsId = defaultPropsId;

        load();
    }

    public CacheConfiguration getCacheConfiguration()
    {
        return this.configuration;
    }

    public Properties getOSCacheProperties()
    {
        return this.oscacheConfiguration;
    }

    public int getTimeToLive()
    {
        return this.timeToLive;
    }

    public String getName()
    {
        return this.name;
    }

    private void load()
    {
        this.oscacheConfiguration = getDefaultConfig();

        try {
            this.oscacheConfiguration = getConfig(this.configuration.getConfigurationId());
        } catch (Exception e) {
            for (Map.Entry<String, Object> entry : this.configuration.entrySet()) {
                if (entry.getKey() == EntryEvictionConfiguration.CONFIGURATIONID) {
                    EntryEvictionConfiguration ec = (EntryEvictionConfiguration) entry.getValue();

                    if (ec.getMode() == EntryEvictionConfiguration.Mode.LRU) {
                        this.oscacheConfiguration.setProperty("cache.algorithm", LRUCache.class.getName());
                        if (ec.containsKey(LRUEvictionConfiguration.MAXENTRIES_ID)) {
                            this.oscacheConfiguration.setProperty("cache.capacity", ec.get(
                                LRUEvictionConfiguration.MAXENTRIES_ID).toString());
                        }
                    }
                }

                this.oscacheConfiguration.setProperty(CacheConfigurationConverter.convert(entry.getKey()), entry
                    .getValue().toString());
            }
        }

        EntryEvictionConfiguration ec =
            (EntryEvictionConfiguration) configuration.get(EntryEvictionConfiguration.CONFIGURATIONID);

        if (ec.getTimeToLive() > 0) {
            timeToLive = ec.getTimeToLive();
        }

        this.name = this.configuration.getConfigurationId();

        if (this.name == null) {
            this.name = UUID.randomUUID().toString();
        }
    }

    private Properties getDefaultConfig()
    {
        Properties defaultConfig = new Properties();

        try {
            loadConfig(defaultConfig, this.defaultPropsId);
        } catch (IOException e) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("Could not load default cache properties", e);
            }
        }

        return defaultConfig;
    }

    private Properties getConfig(String propertiesId) throws IOException
    {
        if (LOG.isInfoEnabled()) {
            LOG.info("Loading cache properties: " + propertiesId);
        }

        Properties props = getDefaultConfig();

        loadConfig(props, propertiesId);

        return props;
    }

    private void loadConfig(Properties props, String propertiesId) throws IOException
    {
        String propertiesFilename = propertiesId + PROPS_EXT;

        File file = new File(PROPS_PATH + propertiesFilename);

        InputStream is = null;

        try {
            if (file.exists()) {
                is = new FileInputStream(file);
            } else {
                is =
                    this.container.getApplicationContext().getResourceAsStream(
                        "/WEB-INF/" + PROPS_PATH + propertiesFilename);
            }

            props.load(is);

            if (LOG.isInfoEnabled()) {
                LOG.info("Properties loaded: " + propertiesFilename);
            }
        } finally {
            if (is != null) {
                try {
                    is.close();
                } catch (IOException e) {
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Failed t close properties file", e);
                    }
                }
            }
        }
    }
}
